/*
  This file is part of the KDE libraries
  Copyright (c) 1999 Preston Brown <pbrown@kde.org>
  Copyright (c) 1997-1999 Matthias Kalle Dalheimer <kalle@kde.org>
  Portions copyright (c) 2002 Tobias Koenig <tokoe82@yahoo.de>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; see the file COPYING.LIB.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <qdir.h>
#include <qdom.h>
#include <qfileinfo.h>
#include <qtextcodec.h>
#include <qtextstream.h>

#include "kconfigbackend.h"
#include "kconfigxmlbackend.h"
#include "kconfigbase.h"
#include "kconfigdata.h"
#include <kglobal.h>
#include <kstandarddirs.h>
#include <ksavefile.h>
#include <kurl.h>

extern bool checkAccess(const QString& pathname, int mode);

bool KConfigXMLBackEnd::parseConfigFiles()
{
    // Parse all desired files from the least to the most specific.
    if (checkAccess(mLocalFileName, W_OK)) 
	mConfigState = KConfigBase::ReadWrite;
    else
	mConfigState = KConfigBase::ReadOnly;

    // Parse the general config files
    if (useKDEGlobals) {
	QStringList kdercs = KGlobal::dirs()->
		findAllResources("config", QString::fromLatin1("kdeglobals.xml"));

	if (checkAccess(QString::fromLatin1("/etc/kderc.xml"), R_OK))
	    kdercs += QString::fromLatin1("/etc/kderc.xml");

        kdercs += KGlobal::dirs()->
	    findAllResources("config", QString::fromLatin1("system.kdeglobals.xml"));

	QStringList::ConstIterator it;

        for (it = kdercs.fromLast(); it != kdercs.end(); --it) {
	    QFile aConfigFile( *it );
	    if (!aConfigFile.open( IO_ReadOnly ))
		continue;
	    parseSingleConfigFile( aConfigFile, 0L, true, (*it != mGlobalFileName) );
	    aConfigFile.close();
	}
    }

    if (!fileName.isEmpty()) {
	QStringList list = KGlobal::dirs()->
		findAllResources(resType, fileName);

	QStringList::ConstIterator it;

	for (it = list.fromLast(); it != list.end(); --it) {
	    QFile aConfigFile( *it );
	    // we can already be sure that this file exists
	    bool bIsLocal = (*it == mLocalFileName);
	    if (aConfigFile.open( IO_ReadOnly )) {
		parseSingleConfigFile( aConfigFile, 0L, false, !bIsLocal );
		aConfigFile.close();
	    }
	}
    }

    return true;
}

void KConfigXMLBackEnd::parseSingleConfigFile(QFile &rFile,
					      KEntryMap *pWriteBackMap,
					      bool bGlobal, bool bDefault)
{
    if (!rFile.isOpen()) // come back, if you have real work for us ;->
	return;
/*
    qDebug("Parsing %s, global = %s default = %s",
	    rFile.name().latin1(), bGlobal ? "true" : "false", bDefault ? "true" : "false");
*/
    QDomDocument doc;

    // check for valid XML header
    if (!doc.setContent(&rFile))
    {
	qDebug("Invalid XML header in %s", rFile.name().latin1());
	return;
    }

    if (doc.doctype().name() != "KDE-Config-File")
    {
	qDebug("The file %s does not contain a valid definition", rFile.name().latin1());
	return;
    }

    QDomElement docElement = doc.documentElement();
    QCString aCurrentGroup;

    bool groupOptionImmutable = false;
    bool groupSkip = false;

    // load all groups
    QDomNodeList groupList = docElement.elementsByTagName("group");
    uint i;
    for (i = 0; i < groupList.count(); ++i)
    {
	QDomElement groupElement = groupList.item(i).toElement();
	aCurrentGroup = groupElement.attribute("name").utf8();	

	/** 
	 * In opposite to the INI-BackEnd, all entries must belong to
	 * a group in the XML-BackEnd. Therefore all entries without a
	 * group are collected in 'Defaults'
	 */
	if (aCurrentGroup == "Defaults")
		aCurrentGroup = "<default>";

	groupOptionImmutable = groupElement.attribute("immutable", "0").toInt();

	KEntryKey groupKey(aCurrentGroup, 0);
	KEntry entry = pConfig->lookupData(groupKey);
	groupSkip = entry.bImmutable;

        if (groupSkip)
	    continue;

	entry.bImmutable = groupOptionImmutable;
	pConfig->putData(groupKey, entry, false);

	if (pWriteBackMap)
	{
	    // add the special group key indicator
            (*pWriteBackMap)[groupKey] = entry;
	}

	bool optionImmutable = false;
	bool optionDeleted = false;
	QString name, locale, value;

	// load all entries of this group
	QDomNodeList entryList = groupElement.elementsByTagName("entry");
	uint j;
	for (j = 0; j < entryList.count(); ++j)
	{
	    QDomElement entryElement = entryList.item(j).toElement();
	    optionImmutable = entryElement.attribute("immutable", QString::number(groupOptionImmutable)).toInt();
	    optionDeleted = entryElement.attribute("deleted", "0").toInt();
	    name = entryElement.attribute("name");
	    value = entryElement.text();
	    locale = entryElement.attribute("locale");
	    if (!locale.isEmpty())
		localeString = locale.utf8();

	    KEntryKey aEntryKey(aCurrentGroup, name.latin1());
	    aEntryKey.bLocal = (!locale.isEmpty());
	    aEntryKey.bDefault = bDefault;

	    KEntry aEntry;
	    aEntry.mValue = value.latin1();
	    aEntry.bGlobal = bGlobal;
	    aEntry.bImmutable = optionImmutable;
	    aEntry.bDeleted = optionDeleted;
	    aEntry.bNLS = (!locale.isEmpty());

	    if (pWriteBackMap) {
		// don't insert into the config object but into the temporary
		// scratchpad map
		pWriteBackMap->insert(aEntryKey, aEntry);
	    } else {
		// directly insert value into config object
		// no need to specify localization; if the key we just
		// retrieved was localized already, no need to localize it again.
		pConfig->putData(aEntryKey, aEntry, false);
	    }
	}
    }
}

void KConfigXMLBackEnd::sync(bool bMerge)
{
    // write-sync is only necessary if there are dirty entries
    if (!pConfig->isDirty())
	return;

    bool bEntriesLeft = true;

    // find out the file to write to (most specific writable file)
    // try local app-specific file first

    if (!fileName.isEmpty()) {
	// Create the containing dir if needed
	if ((resType!="config") && mLocalFileName[0]=='/')
	{
	    KURL path;
	    path.setPath(mLocalFileName);
	    QString dir=path.directory();
	    KStandardDirs::makeDir(dir);
	}

	// Can we allow the write? We can, if the program
	// doesn't run SUID. But if it runs SUID, we must
	// check if the user would be allowed to write if
	// it wasn't SUID.
	if (checkAccess(mLocalFileName, W_OK)) {
	    // is it writable?
	    bEntriesLeft = writeConfigFile( mLocalFileName, false, bMerge );
	}
    }

    // only write out entries to the kdeglobals file if there are any
    // entries marked global (indicated by bEntriesLeft) and
    // the useKDEGlobals flag is set.
    if (bEntriesLeft && useKDEGlobals) {
	// can we allow the write? (see above)
	if (checkAccess ( mGlobalFileName, W_OK )) {
	    writeConfigFile( mGlobalFileName, true, bMerge );
	}
    }
}

static void writeEntries(QDomDocument& doc, QDomElement& groups, const KEntryMap& entryMap, QCString localeString)
{
    QCString currentGroup;
    QDomElement currentGroupElement;
    bool addGroup = false;

    for (KEntryMapConstIterator aIt = entryMap.begin(); aIt != entryMap.end(); ++aIt)
    {
	const KEntryKey &key = aIt.key();

	// Skip default values and group headers.
	if ((key.bDefault) || key.mKey.isEmpty()) 
	    continue; // Skip

	const KEntry &currentEntry = *aIt;

	KEntryMapConstIterator aTestIt = aIt;
	++aTestIt;
	bool hasDefault = (aTestIt != entryMap.end());
	if (hasDefault)
	{
	    const KEntryKey &defaultKey = aTestIt.key();
	    if ((!defaultKey.bDefault) || (defaultKey.mKey != key.mKey) ||
		    (defaultKey.mGroup != key.mGroup) ||
		    (defaultKey.bLocal != key.bLocal))
		hasDefault = false;
	}
 
	if (hasDefault) 
	{
	    // Entry had a default value
	    if ((currentEntry.mValue == (*aTestIt).mValue) && 
	    (currentEntry.bDeleted == (*aTestIt).bDeleted))
	    continue; // Same as default, don't write.
	}
	else
	{
	    // Entry had no default value.
	    if (currentEntry.bDeleted)
		continue; // Don't write deleted entries if there is no default.
	}

	if (currentGroup != key.mGroup) {
	    if (!currentGroup.isEmpty()) {
		groups.appendChild(currentGroupElement);
		addGroup = false;
	    }

	    currentGroup = key.mGroup;
	    currentGroupElement = doc.createElement("group");
	    currentGroupElement.setAttribute("name", (currentGroup == "<default>" ? QCString("Defaults") : currentGroup));
	    addGroup = true;
	}

	QDomElement entryElement = doc.createElement("entry");
	entryElement.setAttribute("name", key.mKey.data());

	// it is data for a group
	if (currentEntry.bDeleted)
	{
	    // Deleted entry
	    entryElement.setAttribute("deleted", true);

	    if ( currentEntry.bNLS )
	    {
		// localized 
		entryElement.setAttribute("locale", localeString);
	    }
	}
	else
	{  
	    if ( currentEntry.bNLS )
	    {
		entryElement.setAttribute("locale", localeString);
		QDomText value = doc.createTextNode(currentEntry.mValue);
		entryElement.appendChild(value);
	    }
	    else
	    {
		// not localized
		QDomText value = doc.createTextNode(currentEntry.mValue);
		entryElement.appendChild(value);
	    }
	}

	currentGroupElement.appendChild(entryElement);
    } // for loop

    if (addGroup)
	groups.appendChild(currentGroupElement);
}

bool KConfigXMLBackEnd::writeConfigFile(QString filename, bool bGlobal,
					bool bMerge)
{
    KEntryMap aTempMap;
    bool bEntriesLeft = false;

    // is the config object read-only?
    if (pConfig->isReadOnly())
	return true; // pretend we wrote it

    QFile rConfigFile( filename );
    if (bMerge)
    {
	// Read entries from disk
	if (rConfigFile.open(IO_ReadOnly))
	{
	    // fill the temporary structure with entries from the file
	    parseSingleConfigFile( rConfigFile, &aTempMap, bGlobal, false );
	    rConfigFile.close();
	}

	KEntryMap aMap = pConfig->internalEntryMap();

	// augment this structure with the dirty entries from the config object
	for (KEntryMapIterator aIt = aMap.begin(); aIt != aMap.end(); ++aIt)
	{
	    const KEntry &currentEntry = *aIt;
	    if (aIt.key().bDefault)
	    {
		aTempMap.replace(aIt.key(), currentEntry);
		continue; 
	    }

	    if (!currentEntry.bDirty)
		continue;

	    // only write back entries that have the same
	    // "globality" as the file
	    if (currentEntry.bGlobal != bGlobal)
	    {
		// wrong "globality" - might have to be saved later
		bEntriesLeft = true;
		continue;
	    }
      
	    // put this entry from the config object into the
	    // temporary map, possibly replacing an existing entry
	    aTempMap.replace(aIt.key(), currentEntry);
	} // loop
    }
    else
    {
	// don't merge, just get the regular entry map and use that.
	aTempMap = pConfig->internalEntryMap();
	bEntriesLeft = true; // maybe not true, but we aren't sure
    }

    // OK now the temporary map should be full of ALL entries.
    // write it out to disk.

    if (!rConfigFile.open(IO_WriteOnly))
    {
	qDebug("can't open '%s' for writing", filename.latin1());
	return false;
    }

    QDomDocument doc("KDE-Config-File");
    doc.appendChild(doc.createProcessingInstruction("xml", "version=\"0.1\" encoding=\"UTF-8\""));

    QDomElement groups = doc.createElement("groups");
    doc.appendChild(groups);

    writeEntries(doc, groups, aTempMap, localeString);

    QTextStream stream(&rConfigFile);
    stream.setEncoding(QTextStream::UnicodeUTF8);
    stream << doc;
    rConfigFile.close();

    return bEntriesLeft;
}
